"""

Copyright (C) 2020-2022 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.exceptions import (
    DirectoryNotFoundError,
    MultipleReposExistError,
    RepoNotFoundError,
    NoReposError,
)
from rse.utils.file import (
    write_json,
    write_file,
    mkdir_p,
    read_json,
    read_file,
    recursive_find,
    get_latest_modified,
)
from rse.main.database.base import Database
from rse.main.parsers import get_parser
from rse.main.parsers.base import ParserBase
from glob import glob
import logging
import shutil
import os
import re

bot = logging.getLogger("rse.main.database.filesystem")


class FileSystemDatabase(Database):
    """A FileSystemDatabase writes raw json to files to a database."""

    database = "filesystem"

    def __init__(self, config_dir, config=None, **kwargs):
        """init for the filesystem ensures that the base folder (named
        according to the studyid) exists.
        """
        self.config = config
        self.create_database(config_dir)

    def create_database(self, config_dir):
        """Create the database. The parent folder must exist."""
        self.data_base = os.path.abspath(os.path.join(config_dir, "database"))
        if not os.path.exists(config_dir):
            raise DirectoryNotFoundError(
                config_dir, "must exist to create database there"
            )
        if not os.path.exists(self.data_base):
            mkdir_p(self.data_base)

    # Global

    def clear(self):
        """clear (delete) all software repositories."""
        for parser_dir in self.iter_parsers(fullpath=True):
            if os.path.exists(parser_dir):
                bot.info(f"Removing {parser_dir}")
                shutil.rmtree(parser_dir)
        return True

    # Get, delete, etc. only require uid

    def exists(self, uid):
        """Determine if a repo exists."""
        try:
            self.get(uid, exact=True)
            return True
        except:
            return False

    def add(self, uid, data=None):
        """Add a new software repository to the database."""
        if uid:
            parser = get_parser(uid, config=self.config)

            if not data:
                data = parser.get_metadata()
            else:
                parser.data = data

            # If it's a parser handoff
            if isinstance(data, ParserBase):
                parser = data
                data = parser.data

            if data:
                bot.info(f"{parser.uid} was added to the the database.")
                return SoftwareRepository(parser, data_base=self.data_base)
        else:
            bot.error("Please define a unique identifier to add.")

    def get_or_create(self, uid):
        """Determine if a repo exists."""
        try:
            repo = self.get(uid, exact=True)
        except:
            repo = self.add(uid)
        return repo

    def get(self, uid=None, exact=False):
        """Get a software repo based on a uid. If exact is not needed, we can
        search for a match based on the partial uid.  If exact is False,
        and a uid is not provided, get the last repository created.
        """
        if not uid and not exact:
            repos = get_latest_modified(self.data_base, pattern="metadata*.json")
            if repos:
                uid = (
                    repos.replace("metadata.json", "")
                    .replace(self.data_base, "")
                    .strip("/")
                )
            if not uid or not repos:
                raise NoReposError

        parser = get_parser(uid, config=self.config)
        return SoftwareRepository(parser, exists=True, data_base=self.data_base)

    def update(self, repo, rewrite=False):
        """Update a repository by retrieving metadata, and then calling update
        on the software repository to save it.
        """
        data = repo.parser.get_metadata()
        if data:
            if rewrite:
                self.add(repo.uid)
            else:
                repo.update(updates=data)

    def label(self, repo, key, value, force=False):
        """Update a repository with a specific key/value pair."""
        if key in repo.data and not force:
            raise RuntimeError(
                f"{key} is already defined for {repo.uid}. Use --force to overwrite."
            )
        bot.debug(f"Adding key {key}:{value}")
        repo.update({key: value})

    def search(self, query, taxonomy=None, criteria=None):
        """A filesystem search can only support returning results with filenames.
        For taxonomy and criteria items, we load them and search.
        We organize results based on the query, taxonomy, and criteria
        The results returned are separate (e.g., a single repo can appear
        in more than one list).
        """
        results = {}

        taxonomy_regex = "(%s)" "|".join(taxonomy or [])
        criteria_regex = "(%s)" "|".join(criteria or [])
        for repo in self.list_repos():

            if query:
                if re.search(query, repo[0], re.IGNORECASE):
                    if query not in results:
                        results[query] = set()
                    results[query].add(repo[0])

            if taxonomy or criteria:
                repo = self.get(repo[0])

            # Add taxonomy items
            if taxonomy:
                for _, tags in repo.load_taxonomy().items():
                    for tag in tags:
                        if re.search(taxonomy_regex, tag, re.IGNORECASE):
                            if tag not in results:
                                results[tag] = set()
                            results[tag].add(repo.uid)

            # Add criteria items
            if criteria:
                for term, annotations in repo.load_criteria().items():
                    for _, annotation in annotations.items():
                        if annotation == "yes" and re.search(
                            criteria_regex, term, re.IGNORECASE
                        ):
                            if term not in results:
                                results[term] = set()
                            results[term].add(repo.uid)

        final = {}
        for key, listing in results.items():
            final[key] = [[x] for x in listing]
        return final

    def delete_repo(self, uid):
        """delete a repo based on a specific identifier."""
        if self.exists(uid):
            repo = self.get(uid)
            os.remove(repo.filename)

            # Remove the directory if no other repos
            dirname = os.path.dirname(repo.filename)
            if not os.listdir(dirname):
                shutil.rmtree(dirname)

            bot.info(f"{uid} has been removed.")
            return True

        bot.error(f"{uid} does not exist in the database.")
        return False

    def delete_parser(self, name):
        """delete all repos for a parser, based on executor's name (str)."""
        parser_dir = os.path.join(self.data_base, name)
        if not os.path.exists(parser_dir):
            bot.info(f"Executor {parser_dir} directory does not exist.")
            return False
        shutil.rmtree(parser_dir)
        return True

    def iter_parsers(self, fullpath=False):
        """list executors based on the subfolders in the base database folder."""
        for contender in os.listdir(self.data_base):
            contender = os.path.join(self.data_base, contender)
            if os.path.isdir(contender):
                if not fullpath:
                    yield os.path.basename(contender)
                else:
                    yield contender

    def list_repos(self, name=None):
        """list software repositories, either under a particular parser name
        or just under all parsers. This returns repos in rows to be printed
        (or otherwise parsed).
        """
        listpath = self.data_base
        if name:
            listpath = os.path.join(listpath, name)
        rows = []
        for filename in recursive_find(listpath, pattern="metadata*.json"):
            rows.append(
                [
                    filename.replace("metadata.json", "")
                    .replace(self.data_base, "")
                    .strip("/")
                ]
            )
        return rows


class SoftwareRepository:
    """A software repository is a filesystem representation of a repo. It can
    take a uid, determine if the repo exists, and then interact with the
    metadata for it. If the repo is instantiated without a unique id
    it is assumed to not exist yet, otherwise it must already
    exist.
    """

    def __init__(self, parser, data_base, exists=False):
        """A SoftwareRepository holds some uid for a parser, and controls
        interaction with the filesystem.

        Arguments:
          parser (str)    : the parser
          data_base (str) : the path where the database exists.
          exists (bool)   : if True, must already exists (default is False)
        """
        self.uid = parser.uid
        self.parser = parser
        self.data_base = data_base
        self.data = {}
        self.criteria = {}
        self.taxonomy = {}
        self.create(exists)

    @property
    def url(self):
        return self.parser.get_url(self.data.get("data", ""))

    @property
    def avatar(self):
        return self.parser.get_avatar(self.data.get("data", ""))

    @property
    def description(self):
        return self.parser.get_description(self.data.get("data", ""))

    @property
    def filename(self):
        return os.path.join(self.parser_dir, "metadata.json")

    @property
    def parser_dir(self):
        return os.path.join(self.data_base, self.parser.uid)

    def update(self, updates=None):
        """Update a data file. This means reading, updating, and writing."""
        updates = updates or {}
        self.data.update(updates)
        self.save()

    def update_criteria(self, uid, username, response):
        """Update a criteria, meaning adding a True/False answer to the
        unique id for the user. We are currently assuming that criteria
        have yes/no responses, and True == yes, False == no.
        """
        if uid not in self.criteria:
            self.criteria[uid] = {}
        if response:
            self.criteria[uid][username] = response

    def create(self, should_exist=False):
        """create the filename if it doesn't exist, otherwise if it should (and
        does not) exit on error.
        """
        if should_exist:
            if not os.path.exists(self.filename):

                # Might be provided prefix
                contenders = glob("%s*" % os.path.join(self.data_base, self.parser.uid))
                if len(contenders) == 1:
                    self.parser.uid = re.sub(
                        "(%s/|[.]json)" % self.data_base,
                        "",
                        contenders[0],
                    )

                elif len(contenders) > 1:
                    raise MultipleReposExistError(self.parser.uid)
                else:
                    raise RepoNotFoundError(self.parser.uid)
            self.data = self.load()
            self.taxonomy = self.load_taxonomy()
            self.criteria = self.load_criteria()

        if not os.path.exists(self.parser_dir):
            mkdir_p(self.parser_dir)

        # If it's the first time saving, create basic file
        if not should_exist:
            self.data = {
                "parser": self.parser.name,
                "uid": self.parser.uid,
                "url": self.parser.get_url(),
                "data": self.parser.export(),
            }
            self.save()

    def export(self):
        """wrapper to expose the executor.export function"""
        return self.parser.export()

    def save(self):
        """Save a json object (metadata.json) for the software repository."""
        write_json(self.data, self.filename)

    def summary(self):
        return self.parser.summary()

    def load(self):
        """Given a software uid, load data from filename."""
        if os.path.exists(self.filename):
            return read_json(self.filename)

    def get_criteria(self):
        """Get loaded criteria"""
        return self.criteria

    def get_taxonomy(self):
        """Get loaded taxonomy"""
        return self.taxonomy

    def load_criteria(self):
        """Given a repository directory, load criteria files if they exist"""
        criteria = {}
        for filename in glob(f"{self.parser_dir}/criteria*.tsv"):
            uid = (
                os.path.basename(filename).replace("criteria-", "").replace(".tsv", "")
            )
            content = read_file(filename)
            if uid not in criteria:
                criteria[uid] = {}
            for row in content:
                row = row.strip()
                if not row:
                    continue
                username, response = row.split("\t")
                criteria[uid][username] = response
        return criteria

    def save_criteria(self):
        """Save criteria to file. Each file is named based on the criteria id,
        and is a tab separated file that includes the username and response.
        """
        for uid, responses in self.criteria.items():
            filename = os.path.join(self.parser_dir, "criteria-%s.tsv" % uid)
            # Sort based on username
            rows = ["%s\t%s" % (k, v) for k, v in sorted(responses.items())]
            write_file(filename, "\n".join(rows))
            bot.debug(f"{uid} saved to {filename}")

    def load_taxonomy(self):
        """Given a repository directory, load taxonomy annotations if they exist
        The taxonomy.tsv file should be a tab separated file with:
        username category-unique-id. This means that we keep a record of
        who has categorized what, and load this information into the
        taxonomy dictionary (organized by the category-unique-id which
        then has a total count and list of users).
        """
        taxonomy = {}
        taxonomy_file = os.path.join(self.parser_dir, "taxonomy.tsv")
        if os.path.exists(taxonomy_file):
            content = read_file(taxonomy_file)
            for row in content:
                row = row.strip()
                if not row:
                    continue
                username, uids = row.split("\t")
                taxonomy[username] = [x.strip() for x in uids.split(",")]
        return taxonomy

    def save_taxonomy(self):
        """Save taxonomy to file. Each file is named taxonomy.tsv,
        and is a tab separated file that includes the username and response.
        """
        filename = os.path.join(self.parser_dir, "taxonomy.tsv")
        rows = ["%s\t%s" % (k, ",".join(v)) for k, v in sorted(self.taxonomy.items())]
        write_file(filename, "\n".join(rows))
        bot.debug(f"{self.uid} saved to {filename}")

    # Annotation

    def has_criteria_annotation(self, uid, username):
        """Determine if a repository has been annotated by a user."""
        if uid not in self.criteria:
            return False
        if username not in self.criteria[uid]:
            return False
        return True

    def has_taxonomy_annotation(self, username):
        """Determine if a repository has been annotated by a user."""
        if username not in self.taxonomy:
            return False
        return True



"""

Copyright (C) 2020-2022 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from flask_socketio import SocketIO
from flask import Flask

from rse.logger import bot
from rse.defaults import RSE_HOSTNAME

import logging
import os


class ResearchSoftwareServer(Flask):
    def __init__(self, *args, **kwargs):
        super(ResearchSoftwareServer, self).__init__(*args, **kwargs)


app = ResearchSoftwareServer(__name__)
app.config.from_object("rse.app.config")

# turn the flask app into a socketio app
socketio = SocketIO(app, async_mode=None, logger=True, engineio_logger=True)

from rse.app.views import *
from rse.app.api import *


def start(
    port=5000, debug=True, client=None, host=None, level="DEBUG", disable_annotate=False
):
    """Start can be invoked when this file is executed (see __main__ below)
    or used as a function to programmatically start a server. If started
    via rse start, we can add the encyclopedia client to the server.
    If you want to change the hostname, set the environment variable
    RSE_HOSTNAME or set on command line with rse start.
    """
    host = host or RSE_HOSTNAME
    bot.info(f"Research Software Encyclopedia: running on http://{host}:{port}")

    # If the user doesn't specify a queue, use default
    if not client:
        from rse.main import Encyclopedia

        client = Encyclopedia()

    # Customize the logger to log to the app folder
    file_handler = logging.FileHandler(os.path.join(client.config_dir, "dashboard.log"))
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    file_handler.setFormatter(formatter)
    app.logger.addHandler(file_handler)
    app.logger.setLevel(getattr(logging, level))

    # Add the queue to the server
    app.client = client
    app.disable_annotate = disable_annotate
    socketio.run(app, port=port, debug=debug, host=host)


# This is how the command line version will run
if __name__ == "__main__":
    start()



"""

Copyright (C) 2020-2022 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import logging
import random
import requests
from time import sleep
from rse.utils.urls import get_user_agent, check_response

from .base import ParserBase

bot = logging.getLogger("rse.main.parsers.github")


class GitHubParser(ParserBase):

    name = "github"
    matchstring = "github"

    def __init__(self, uid=None, **kwargs):
        super().__init__(uid)

    def _set_uid(self, uid):
        """Given some kind of GitHub url, parse the uid"""
        uid = uid.replace(":", "/")
        owner, repo = uid.replace(".git", "").split("/")[-2:]
        return "{}/{}".format(owner, repo)

    def load_secrets(self):
        """load secrets, namely the GitHub token"""
        self.token = self.get_setting("TOKEN")

    def get_url(self, data=None):
        """a common function for a parser to return the html url for the
        upper level of metadata
        """
        data = data or self.data
        return data.get("html_url")

    def get_avatar(self, data=None):
        """a common function for a parser to return an image."""
        data = data or self.data
        return data.get("owner", {}).get("avatar_url", "")

    def get_description(self, data=None):
        """a common function for a parser to return a description."""
        data = data or self.data
        return data.get("description")

    def get_org_repos(self, org, paginate=True, delay=None):
        """
        A helper function to get a listing of org repos.
        """
        self.load_secrets()
        url = "https://api.github.com/orgs/%s/repos?per_page=100" % (org)
        headers = {
            "Accept": "application/vnd.github.symmetra-preview+json",
            "User-Agent": get_user_agent(),
        }
        if self.token:
            headers["Authorization"] = "token %s" % self.token

        repos = []

        # Start at 2, as 1 is implied to be the first
        page = 2
        original_url = url
        while url is not None:
            response = requests.get(url, headers=headers)
            data = check_response(response)

            # Reset the url to be None
            url = None
            if data and paginate:
                url = original_url + "&page=%s" % page

            repos = repos + data
            page += 1
            # Sleep for a random amount of time to give a rest!
            sleep(delay or random.choice(range(1, 10)) * 0.1)
        return repos

    def get_metadata(self, uri=None):
        """Retrieve repository metadata. The common metadata (timestamp) is
        added by the software repository parser, and here we need to
        ensure that the url field is populated with a correct url.

        Arguments:
        uri (str) : a repository uri string to override one currently set
        """
        if uri:
            self.set_uri(uri)
        self.load_secrets()
        repo = "/".join(self.uid.split("/")[-2:])
        url = "https://api.github.com/repos/%s" % (repo)
        headers = {
            "Accept": "application/vnd.github.symmetra-preview+json",
        }
        if self.token:
            headers["Authorization"] = "token %s" % self.token

        response = requests.get(url, headers=headers)

        # Successful query!
        data = check_response(response)
        if data is None:
            return None

        self.data = self.parse_github_repo(data)
        return self.data

    def parse_github_repo(self, repo):
        """
        Given an API response for a GitHub repository, parse a minimal set.
        """
        self.load_secrets()
        headers = {
            "Accept": "application/vnd.github.symmetra-preview+json",
        }
        if self.token:
            headers["Authorization"] = "token %s" % self.token

        # Only save minimal set
        data = {}
        for key in [
            "name",
            "url",
            "full_name",
            "html_url",
            "private",
            "description",
            "created_at",
            "updated_at",
            "clone_url",
            "homepage",
            "size",
            "stargazers_count",
            "watchers_count",
            "language",
            "open_issues_count",
            "license",
            "subscribers_count",
        ]:
            if key in repo:
                data[key] = repo[key]
        data["owner"] = {}
        for key in ["html_url", "avatar_url", "login", "type"]:
            data["owner"][key] = repo["owner"][key]

        # Also try to get topics
        headers.update({"Accept": "application/vnd.github.mercy-preview+json"})
        url = "%s/topics" % repo["url"]
        response = requests.get(url, headers=headers)

        # Add topics on successful query
        topics = check_response(response)
        if topics is not None:
            data["topics"] = topics.get("names", [])

        # Add topics from another source
        if "topics" not in data and "topics" in repo:
            data["topics"] = repo["topics"]
        elif "topics" in data and "topics" in repo:
            data["topics"] += [x for x in repo["topics"] if x not in data["topics"]]

        return data



"""

Copyright (C) 2020-2022 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.main import Encyclopedia


def main(args, extra):

    # Create a queue object, run the command to match to an executor
    enc = Encyclopedia(config_file=args.config_file, database=args.database)

    # If a file is provided:
    if args.file:
        enc.bulk_update(args.file, rewrite=args.rewrite)
    else:
        enc.update(args.uid, rewrite=args.rewrite)



"""

Copyright (C) 2022 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.main.parsers import GitHubParser
import logging
import requests

from .base import ScraperBase

bot = logging.getLogger("rse.main.scrapers.ropensci")


class ROpenSciScraper(ScraperBase):

    name = "ropensci"

    def __init__(self, query=None, **kwargs):
        super().__init__(query)

    def latest(self, paginate=False, delay=0.0):
        """
        Only return latest (top page) of results (default is sort by created)
        """
        return self.scrape(paginate=paginate, delay=delay)

    def search(self, query=None, paginate=True, delay=0.0):
        """
        Return all paginated results with search (no query accepted for now)
        """
        return self.scrape(paginate=paginate, delay=delay)

    def read_registry(self):
        """
        Read the registry file to create a lookup of repos based on GitHub URL.
        """
        lookup = {}
        response = requests.get(
            "https://raw.githubusercontent.com/ropensci/roregistry/gh-pages/registry.json"
        ).json()
        for entry in response.get("packages", []):
            # This is the GitHub URL
            url = entry.get("github")
            if not url:
                continue
            lookup[url] = entry
        return lookup

    def get_registry_topics(self, meta):
        """
        Given a metadata entry from the registry lookup, parse and return topics
        """
        topics = [x.strip() for x in meta.get("keywords", "").split(",") if x.strip()]
        if meta.get("ropensci_category"):
            topics += [meta.get("ropensci_category")]
        return topics

    def scrape(self, paginate=False, delay=None):
        """A shared function to scrape a set of repositories. Since the JoSS
        pages for a search and the base are the same, we can use a shared
        function.
        """
        # This serves as 1: a lookup, and 2: to find non ropensci repos!
        lookup = self.read_registry()

        # Full GitHub urls that we expect to find!
        names = set(lookup.keys())

        parser = GitHubParser()
        repos = parser.get_org_repos("ropensci", paginate=paginate, delay=delay)

        for entry in repos:

            # We determine belonging based on the github url
            if entry["html_url"] not in names:
                bot.info("Skipping repository: %s" % entry["html_url"])
                continue

            meta = lookup[entry["html_url"]]

            # Get topics from R metadata
            topics = self.get_registry_topics(meta)
            names.remove(entry["html_url"])
            if "topics" not in entry:
                entry["topics"] = []
            entry["topics"] += topics

            # Add a description if missing
            if not entry.get("description") and meta.get("description"):
                entry["description"] = meta["description"]

            bot.info("Adding repository: %s" % entry["html_url"])
            self.results.append(parser.parse_github_repo(entry))

        # If paginate is True, we intend to add ALL repos, so check those still remaining
        # E.g., there are repos in other orgs that won't be found above
        if paginate and names:
            for name in names:

                parser = GitHubParser(uid=name)
                # Topics will be added here!
                entry = parser.get_metadata()
                meta = lookup[name]

                # Add a description if missing
                if not entry.get("description") and meta.get("description"):
                    entry["description"] = meta["description"]
                entry["topics"] += self.get_registry_topics(meta)
                bot.info("Adding repository: %s" % entry["html_url"])
                self.results.append(entry)

        return self.results

    def create(self, database=None, config_file=None):
        """After a scrape (whether we obtain latest or a search query) we
        run create to create software repositories based on results.
        """
        from rse.main import Encyclopedia

        client = Encyclopedia(config_file=config_file, database=database)
        for result in self.results:
            uid = result["html_url"].split("//")[-1]

            # Add results that don't exist
            if not client.exists(uid):
                client.add(uid, data=result)



"""

Copyright (C) 2020-2022 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.main import Encyclopedia


def main(args, extra):

    # Create an encyclopedia object
    enc = Encyclopedia(config_file=args.config_file, database=args.database)

    # Does the user want to search for repos with one or more topics?
    if not args.search:
        topics = enc.topics(pattern=args.pattern)
        print("\n".join(topics))
    else:
        repos = enc.repos_by_topics(topics=args.search)
        print("\n".join(repos))



"""

Copyright (C) 2022 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.utils.urls import get_user_agent
from rse.main.parsers import get_parser
import logging
import requests
import random
import sys
from time import sleep

from .base import ScraperBase

bot = logging.getLogger("rse.main.scrapers.molssi")


class MolssiScraper(ScraperBase):

    name = "molssi"

    def __init__(self, query=None, **kwargs):
        super().__init__(query)
        self.baseurl = "https://api.molssi.org"

    def latest(self, paginate=False, delay=0.0):
        return self.scrape(paginate=paginate, delay=delay)

    def search(self, query, paginate=True, delay=0.0):
        return self.scrape(paginate=paginate, delay=delay, query=query)

    def scrape(self, paginate=False, delay=None, query=""):
        """A shared function to scrape software from molssi."""
        url = "%s/search" % self.baseurl
        try:
            from bs4 import BeautifulSoup
            import bs4
        except ImportError:
            sys.exit("BeautifulSoup is required. pip install rse[scraper].")

        data = {
            "domain": "",
            "languages": "[]",
            "mm_filters": None,
            "price": "",
            "qm_filters": "{}",
            "query_text": query,
        }

        response = requests.get(
            url, params=data, headers={"User-Agent": get_user_agent()}
        )
        soup = BeautifulSoup(response.text, "html.parser")
        contenders = soup.find_all("a", {"class": "card-link"}, href=True)
        for contender in contenders:
            href = contender.attrs.get("href")
            if "software_detail" not in href:
                continue

            href = "%s/%s" % (self.baseurl, href)

            # Sleep for a random amount of time to give a rest!
            sleep(delay or random.choice(range(1, 10)) * 0.1)
            response = requests.get(href, headers={"User-Agent": get_user_agent()})
            meta_soup = BeautifulSoup(response.text, "html5lib")
            links = meta_soup.find_all("a", href=True)

            source_code = None
            citation = None
            for link in links:
                if "source code" in link.text.lower():
                    source_code = link.attrs.get("href")

                    if (
                        source_code
                        and "github" in source_code
                        or "gitlab" in source_code
                    ):
                        break
                    if (
                        source_code
                        and "github" not in source_code
                        or "gitlab" not in source_code
                    ):
                        source_code = None

            # If no source code, continue parsing until we find
            if not source_code:
                for link in links:
                    href = link.attrs.get("href")
                    if href and "github" in href or "gitlab" in href:
                        source_code = href
                        break

            # We couldn't find the source code
            if not source_code:
                continue

            # Do we have a citation?
            for bolded in meta_soup.find_all("b"):
                if "citation" in bolded.text.lower():
                    if "row" in bolded.parent.parent.attrs.get("class"):
                        nexts = list(bolded.parent.parent.parent.next_elements)
                        for next_element in nexts:

                            # The citation is the first link under this bolded section
                            if isinstance(
                                next_element, bs4.element.Tag
                            ) and next_element.find_next("a"):
                                citation = next_element.find_next("a")
                                citation = citation.attrs.get("href")
                                break

            # These are not citations!
            if citation and ("mailto" in citation or citation == source_code):
                citation = None

            repo = None
            if source_code and citation:
                bot.info("Found repository %s and doi %s" % (source_code, citation))
                repo = {"url": source_code, "doi": citation}
            elif source_code:
                bot.info("Found repository %s" % source_code)
                repo = {"url": source_code}

            if repo:
                self.results.append(repo)

        return self.results

    def create(self, database=None, config_file=None):
        """After a scrape (whether we obtain latest or a search query) we
        run create to create software repositories based on results.
        """
        from rse.main import Encyclopedia

        client = Encyclopedia(config_file=config_file, database=database)
        for result in self.results:
            uid = result["url"].split("//")[-1]

            # If a repository is added that isn't represented
            try:
                repo = get_parser(uid)
            except NotImplementedError as exc:
                bot.warning(exc)
                continue

            # Add results that don't exist
            if not client.exists(repo.uid):
                client.add(repo.uid)
                client.label(repo.uid, key="doi", value=result.get("doi"))
